<?php
namespace Teknisa\DF\Services\Emissao;

use Financeiro\Entities\Boletreceber;
use Financeiro\Entities\Instcobranca;
use Fiscal\Entities\Notafiscal;
use Geral\Entities\Endeclie;
use Teknisa\DF\Services\Emissao\Banco\GeraEmissaoBoletoSantanderService;
use Teknisa\DF\Services\Emissao\Banco\GeraEmissaoBoletoBanrisulService;
use Teknisa\DF\Services\Emissao\Banco\GeraEmissaoBoletoBrasilService;
use Teknisa\DF\Services\Emissao\Banco\GeraEmissaoBoletoCaixaService;
use Teknisa\DF\Services\Emissao\Banco\GeraEmissaoBoletoBradescoService;
use Teknisa\DF\Services\Emissao\Banco\GeraEmissaoBoletoItauService;

/**
 * Class GeraEmissaoBoletoService
 * @package Teknisa\DF\Services\Emissao
 */
abstract class GeraEmissaoBoletoService
{
    const SEQUENCIA_NOSSO_NUMERO       = 1;
    const SEQUENCIA_CODIGO_BARRAS      = 2;
    const MODULO_DIGITO_VERIFICADOR    = 11;
    const TAMANHO_NOSSO_NUMERO         = 12;
    const PREFIXO_CDCONTADOR           = "NOSSONUMERO";
    const CODIGO_COMPENSACAO_SANTANDER = "033";
    const CODIGO_COMPENSACAO_BRADESCO  = "237";
    const CODIGO_COMPENSACAO_ITAU      = "341";
    const CODIGO_COMPENSACAO_CAIXA     = "104";
    const CODIGO_COMPENSACAO_BRASIL    = "001";
    const CODIGO_COMPENSACAO_BANRISUL  = "041";
    const MOEDA_REAL                   = "9";
    const OUTRA_MOEDA                  = "8";

    protected $boleto;
    protected $digitoCodigoBarras;
    protected $digitoCodigoBarrasBrasil;

    private $codigoReduzido;
    private $carteira;
    private $cdcedenteban;
    private $nossoNumero;
    private $valorBoleto;
    private $cdbanccnab;
    private $dataVencimento;
	private $cdagencia;
	private $cdcontcorr;
    private $digitoVerificadorCedenteCaixa;
    private $cliente;
    private $vrmultabrec;
    private $endeclie;
    private $estadoCliente;
    private $municipioCliente;
    private $cepCliente;
    private $cdfilial;
    private $nrlanbolrec;
    private $dsendeclie;
    private $nmepresa;
    private $nrinsjurempr;
    private $nrnotafisc;
    private $instrucao1;
    private $instrucao2;

    /**
     * @param Boletreceber $boleto
     */
    public function __construct(Boletreceber $boleto)
    {
        /** @var Boletreceber boleto */
        $this->boleto = $boleto;

        $this->digitoCodigoBarras = null;
        $this->setCodigoReduzido($this->boleto->getCdreducont());
        $this->setCarteira($this->boleto->getCdcartcobbol());
        $this->setCdcedenteban($this->boleto->getReducont()->getCdcedenteban());
        $this->setNossoNumero($this->boleto->getCdnosnumrec());
        $this->setValorBoleto($this->boleto->getVrfinalbrec());
        $this->setCdbanccnab($this->boleto->getReducont()->getAgencia()->getBanco()->getCdbanccnab());
        $this->setDataVencimento($this->boleto->getDtvenbolrec());
		$this->setCdagencia($this->boleto->getReducont()->getAgencia()->getCdAgencia());
		$this->setCdcontcorr($this->boleto->getReducont()->getCdcontcorr());
        $this->setDigitoVerificadorCedenteCaixa($this->boleto->getReducont()->getCdcedenteban());
        $this->setCliente($this->boleto->getCliente()->getNmrazsocclie());
        $this->setVrmultabrec($this->boleto->getVrmultabrec());
        $this->setEndeclie($this->boleto->getCliente()->getEndeclieprinc());
        $this->setDsendeclie($this->getEndeclie()->getDsendeclie());
        $this->setEstadoCliente($this->getEndeclie()->getSgestado());
        $this->setMunicipioCliente($this->getEndeclie()->getMunicipio()->getNmmunicipio());
        $this->setCepCliente($this->getEndeclie()->getNrcepclie());
        $this->setCdfilial($this->boleto->getCdfilial());
        $this->setNrlanbolrec($this->boleto->getNrlanbolrec());
        $this->setNmepresa($this->boleto->getEmpresa()->getNmempresa());
        $this->setNrinsjurempr($this->boleto->getEmpresa()->getNrinsjurempr());
        $this->setNrnotafisc();
        $this->setInstrucao1($this->boleto->getCdins1cobbol());
        $this->setInstrucao2($this->boleto->getCdins2cobbol());
    }

    /**
     * @param Boletreceber $boleto
     * @return null
     */
    public static function factory(Boletreceber $boleto)
    {
        $codigoCompensacao = $boleto->getReducont()->getAgencia()->getBanco()->getCdbanccnab();
        if ($codigoCompensacao === self::CODIGO_COMPENSACAO_SANTANDER) {
            return new GeraEmissaoBoletoSantanderService($boleto);
        }
		if ($codigoCompensacao === self::CODIGO_COMPENSACAO_BRADESCO) {
			return new GeraEmissaoBoletoBradescoService($boleto);
        }
        if ($codigoCompensacao === self::CODIGO_COMPENSACAO_BANRISUL) {
            return new GeraEmissaoBoletoBanrisulService($boleto);
        }
		if ($codigoCompensacao === self::CODIGO_COMPENSACAO_ITAU) {
            return new GeraEmissaoBoletoItauService($boleto);
        }
		if ($codigoCompensacao === self::CODIGO_COMPENSACAO_CAIXA) {
            return new GeraEmissaoBoletoCaixaService($boleto);
        }
		if ($codigoCompensacao === self::CODIGO_COMPENSACAO_BRASIL) {
            return new GeraEmissaoBoletoBrasilService($boleto);
        }

        return null;
    }

    public function setDigitoCodigoBarras($digitoCodigoBarras)
    {
        $this->digitoCodigoBarras = $digitoCodigoBarras;
    }

    public function getDigitoCodigoBarras()
    {
        if ($this->digitoCodigoBarras === null) 
		{
            throw new \Exception("[ERRO_DESENVOLVEDOR] O digito do codigo de barras ainda nao foi calculado.");
        }

        return $this->digitoCodigoBarras;
    }

    /**
     * @return null
     * @throws \Exception
     */
    public function getDigitoCodigoBarrasBrasil()
    {
        if ($this->digitoCodigoBarras === null)
        {
            throw new \Exception("[ERRO_DESENVOLVEDOR] O digito do codigo de barras ainda nao foi calculado.");
        }

        return $this->digitoCodigoBarras;
    }

    /**
     * @param mixed $digitoCodigoBarrasBrasil
     */
    public function setDigitoCodigoBarrasBrasil($digitoCodigoBarrasBrasil)
    {
        $this->digitoCodigoBarrasBrasil = $digitoCodigoBarrasBrasil;
    }


    protected function getDigitoVerificador($sequencia, $tipoSequencia = self::SEQUENCIA_CODIGO_BARRAS)
    {
        // Inverte o nosso numero.
        $sequencia   = strrev($sequencia);
        $peso = 2;
        $acumulador    = 0;

        // Este loop percorrera todos os caracteres da variavel $nossonumero
        for ($i = 0; $i < strlen($sequencia); $i++) {
            $acumulador += ((int)substr($sequencia, $i, 1) * $peso);
            $peso = $this->incrementaValorPeso($peso, 2, 9);
        }

        $digito = $acumulador % self::MODULO_DIGITO_VERIFICADOR;
        if ($tipoSequencia === self::SEQUENCIA_CODIGO_BARRAS) {
            $digito = $this->avaliarDigitoRetornoCodigoBarras($digito);
        } else {
            $digito = $this->avaliarDigitoRetornoNossoNumero($digito);
        }

        $this->setDigitoCodigoBarras($digito);

        return $digito;
    }

    protected function getDigitoVerificadorCaixa($sequencia)
    {
        // Inverte o nosso numero.
        $sequencia   = strrev($sequencia);
        $peso = 2;
        $acumulador    = 0;

        // Este loop percorrera todos os caracteres da variavel $nossonumero
        for ($i = 0; $i < strlen($sequencia); $i++) {
            $acumulador += ((int)substr($sequencia, $i, 1) * $peso);
            $peso = $this->incrementaValorPeso($peso, 2, 9);
        }

        $digito = 11 - ($acumulador % self::MODULO_DIGITO_VERIFICADOR);
        $digito = $digito == 0 || $digito > 9 ? 1 : $digito;

        $this->setDigitoCodigoBarras($digito);

        return $digito;
    }

    protected function getDigitoVerificadorBrasil($sequencia)
    {
        // Inverte o nosso numero.
        $sequencia   = strrev($sequencia);
        $peso = 9;
        $acumulador    = 0;

        // Este loop percorrera todos os caracteres da variavel $nossonumero
        for ($i = 0; $i < strlen($sequencia); $i++) {
            $acumulador += ((int)substr($sequencia, $i, 1) * $peso);
            $peso = $this->incrementaValorPesoBrasil($peso, 2, 9);
        }

        $digito = $acumulador % self::MODULO_DIGITO_VERIFICADOR;
        $digito = $this->avaliarDigitoRetornoCodigoBarrasBrasil($digito);

        $this->setDigitoCodigoBarrasBrasil($digito);
        return $digito;
    }
	
	
	
    /**
     * Verificar as regras de criacao do digito verificador. No caso deste banco, caso o valor do digito passado por
     * parametro for igual a 10, este metodo retornara 1. Caso o valor seja 1 ou 0, este retornara 0. Caso o valor seja
     * qualquer outro diferente dos valores anteriores, o valor retornado sera a subtracao do modulo do digito
     * verificador pelo valor passado no parametro.
     *
     * @param $digito
     * @return int
     */
    private function avaliarDigitoRetornoNossoNumero($digito)
    {
        if ($digito === 10) {
            return 1;
        } else if ($digito < 2) {
            return 0;
        } else {
            return self::MODULO_DIGITO_VERIFICADOR - $digito;
        }
    }

    private function avaliarDigitoRetornoCodigoBarras($digito)
    {
        if ($digito < 2 || $digito === 10) {
            return 1;
        } else {
            return self::MODULO_DIGITO_VERIFICADOR - $digito;
        }

//        return $this->avaliarDigitoRetornoNossoNumero($digito);
    }

    private function avaliarDigitoRetornoCodigoBarrasBrasil($digito)
    {
        if ($digito === 10) {
            return 'X';
        } else {
            return $digito;
        }
    }
	


    /** Incrementa o peso utilizado para calcular o digito verificador. Este metodo adicionara 1 ao valor do peso ate
     * que este valor supere o valor maximo permitido como peso do verificador. Quando isto ocorrer, o valor de peso
     * retornará ao valor minimo.
     *
     * @param $peso
     * @param $pesoMin
     * @param $pesoMax
     * @return int
     */
    protected function incrementaValorPeso($peso, $pesoMin, $pesoMax)
    {
        $peso++;
        if ($peso > $pesoMax) {
            $peso = $pesoMin;
        }

        return $peso;


    }

    protected function incrementaValorPesoBrasil($peso, $pesoMin, $pesoMax){
        $peso--;
        if ($peso < $pesoMin) {
            $peso = $pesoMax;
        }

        return $peso;
    }


    // ATRIBUTOS DO BOLETO EMITIDO.
    /**
     * Atributos da classe.
     *
     * @param $carteira
     */
    protected function setCarteira($carteira)
    {
        $this->carteira = $carteira;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getCarteira()
    {
        return $this->carteira;
    }
	
	protected function setCdcontcorr($cdcontcorr)
    {
        $this->cdcontcorr = $cdcontcorr;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getCdcontcorr()
    {
        return $this->cdcontcorr;
    }

    /**
     * Atributos da classe.
     *
     * @param $codigoCedente
     */
    protected function setCdcedenteban($codigoCedente)
    {
        $this->cdcedenteban = $codigoCedente;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getCdcedenteban()
    {
        return $this->cdcedenteban;
    }
	
	protected function setCdAgencia($codigoAgencia)
    {
        $this->cdagencia = $codigoAgencia;
    }
	
	protected function getCdAgencia()
    {
        return $this->cdagencia;
    }
	

    /**
     * Atributos da classe.
     *
     * @param $codigoReduzido
     */
    protected function setCodigoReduzido($codigoReduzido)
    {
        $this->codigoReduzido = $codigoReduzido;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getCodigoReduzido()
    {
        return $this->codigoReduzido;
    }

    /**
     * Atributos da classe.
     *
     * @param $dataVencimento
     */
    protected function setDataVencimento($dataVencimento)
    {
        $this->dataVencimento = $dataVencimento;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getDataVencimento()
    {
        return $this->dataVencimento;
    }

    /**
     * Atributos da classe.
     *
     * @param $identificadorBanco
     */
    protected function setCdbanccnab($identificadorBanco)
    {
        $this->cdbanccnab = $identificadorBanco;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getCdbanccnab()
    {
        return $this->cdbanccnab;
    }

    /**
     * Atributos da classe.
     *
     * @param $nossoNumero
     */
    protected function setNossoNumero($nossoNumero)
    {
        $this->nossoNumero = $nossoNumero;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getNossoNumero()
    {
        return $this->nossoNumero;
    }


    /**
     * @return mixed
     */
    public function getCliente()
    {
        return $this->cliente;
    }

    /**
     * @param mixed $cliente
     */
    public function setCliente($cliente)
    {
        $this->cliente = $cliente;
    }


    /**
     * Atributos da classe.
     *
     * @param $valorBoleto
     */
    protected function setValorBoleto($valorBoleto)
    {
        $this->valorBoleto = $valorBoleto;
    }

    /**
     * Atributos da classe.
     *
     * @return mixed
     */
    protected function getValorBoleto()
    {
        return $this->valorBoleto;
    }

    /**
     * @return mixed
     */
    protected function getDigitoVerificadorCedenteCaixa()
    {
        return $this->digitoVerificadorCedenteCaixa;
    }

    /**
     * @param mixed $digitoVerificadorCedenteCaixa
     */
    protected function setDigitoVerificadorCedenteCaixa($digitoVerificadorCedenteCaixa)
    {
        // Inverte o nosso numero.
        $sequencia   = strrev($digitoVerificadorCedenteCaixa);
        $peso = 2;
        $acumulador    = 0;

        // Este loop percorrera todos os caracteres da variavel $nossonumero
        for ($i = 0; $i < strlen($sequencia); $i++) {
            $acumulador += ((int)substr($sequencia, $i, 1) * $peso);
            $peso = $this->incrementaValorPeso($peso, 2, 9);
        }

        $acumulador = $acumulador % self::MODULO_DIGITO_VERIFICADOR;
        $acumulador = self::MODULO_DIGITO_VERIFICADOR - $acumulador;
        $acumulador = $acumulador > 9 ? 0 : $acumulador;

        $this->digitoVerificadorCedenteCaixa = $acumulador;
    }


    /**
     * @return mixed
     */
    public function getVrmultabrec()
    {
        return $this->vrmultabrec;
    }

    /**
     * @param mixed $vrmultabrec
     */
    public function setVrmultabrec($vrmultabrec)
    {
        $this->vrmultabrec = $vrmultabrec;
    }

    /**
     * @return Endeclie
     */
    public function getEndeclie()
    {
        return $this->endeclie;
    }

    /**
     * @param Endeclie $endeclie
     */
    public function setEndeclie($endeclie)
    {
        $this->endeclie = $endeclie;
    }

    /**
     * @return mixed
     */
    public function getEstadoCliente()
    {
        return $this->estadoCliente;
    }

    /**
     * @param mixed $bairroCliente
     */
    public function setEstadoCliente($bairroCliente)
    {
        $this->estadoCliente = $bairroCliente;
    }

    /**
     * @return mixed
     */
    public function getMunicipioCliente()
    {
        return $this->municipioCliente;
    }

    /**
     * @param mixed $municipioCliente
     */
    public function setMunicipioCliente($municipioCliente)
    {
        $this->municipioCliente = $municipioCliente;
    }

    /**
     * @return mixed
     */
    public function getCepCliente()
    {
        return $this->cepCliente;
    }

    /**
     * @param mixed $cepCliente
     */
    public function setCepCliente($cepCliente)
    {
        $this->cepCliente = $cepCliente;
    }

    /**
     * @return mixed
     */
    public function getCdfilial()
    {
        return $this->cdfilial;
    }

    /**
     * @param mixed $cdfilial
     */
    public function setCdfilial($cdfilial)
    {
        $this->cdfilial = $cdfilial;
    }

    /**
     * @return mixed
     */
    public function getNrlanbolrec()
    {
        return $this->nrlanbolrec;
    }

    /**
     * @param mixed $nrlanbolrec
     */
    public function setNrlanbolrec($nrlanbolrec)
    {
        $this->nrlanbolrec = $nrlanbolrec;
    }

    /**
     * @return mixed
     */
    public function getDsendeclie()
    {
        return $this->dsendeclie;
    }

    /**
     * @param mixed $dsendeclie
     */
    public function setDsendeclie($dsendeclie)
    {
        $this->dsendeclie = $dsendeclie;
    }

    /**
     * @return mixed
     */
    public function getNmepresa()
    {
        return $this->nmepresa;
    }

    /**
     * @param mixed $nmepresa
     */
    public function setNmepresa($nmepresa)
    {
        $this->nmepresa = $nmepresa;
    }

    /**
     * @return mixed
     */
    public function getNrinsjurempr()
    {
        return $this->nrinsjurempr;
    }

    /**
     * @param mixed $nrinsjurempr
     */
    public function setNrinsjurempr($nrinsjurempr)
    {
        $this->nrinsjurempr = $nrinsjurempr;
    }

    /**
     * @return mixed
     */
    public function getNrnotafisc()
    {
        return $this->nrnotafisc;
    }

    /**
     * @param mixed
     */
    public function setNrnotafisc()
    {
        /** @var $notafiscal /Repositories/Fiscal/Notafiscal */
        $notafiscal = Notafiscal::getRepository();
        $nrnotafisc = $notafiscal->getMaxNrnotafisc($this->getCdfilial(), $this->getNrlanbolrec());

        $this->nrnotafisc = $nrnotafisc[0]['NRNOTAFISC'];
    }

    /**
     * @return mixed
     */
    public function getInstrucao1()
    {
        return $this->instrucao1;
    }

    /**
     * @param mixed $instrucao1
     */
    public function setInstrucao1($instrucao1)
    {
        if($instrucao1) {
            /** @var Instcobranca $instrucao */
            $instrucao = Instcobranca::findOneByCdinstcobr($instrucao1);
            if ($instrucao)
                $instrucao1 = $instrucao->getNminstcobr();
            else
                $instrucao1 = "";
        }
        else
            $instrucao1 = "";

        $this->instrucao1 = $instrucao1;
    }

    /**
     * @return mixed
     */
    public function getInstrucao2()
    {
        return $this->instrucao2;
    }

    /**
     * @param mixed $instrucao2
     */
    public function setInstrucao2($instrucao2)
    {
        if($instrucao2) {
            /** @var Instcobranca $instrucao */
            $instrucao = Instcobranca::findOneByCdinstcobr($instrucao2);
            $instrucao2 = $instrucao?$instrucao->getNminstcobr():"";
        }

        $this->instrucao2 = $instrucao2;
    }

    public abstract function getNossoNumeroComDigito();

    public abstract function getLinhaDigitavel($nossoNumero);

    public abstract function getCodigoDeBarras($nossoNumero);

    public abstract function getRelatorioBoleto($nrorg, $cdreducont);

    public abstract function chamaRelatorio();

}
